/*

  mmapc.c  -- C code for mmap VxD
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/


#include <string.h>
#pragma intrinsic(strlen, strcpy, memset, memcpy)


#define WANTVXDWRAPS
#include <basedef.h>
#include <vmm.h>
#include <vxdwraps.h>

#include "vmmextra.h"

static char copyright[] = "Copyright 1997 Michael Callahan\n";

typedef struct MEMORY_BLOCK {
	DWORD	mb_addr;
	DWORD	mb_size;
} MEMORY_BLOCK, *PMEMORY_BLOCK;


#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static DWORD cbmyVMarea = 0;

#pragma VxD_INIT_DATA_SEG
#pragma VxD_INIT_CODE_SEG

#pragma warning(disable:4035)
DWORD __declspec(naked) _Allocate_Device_CB_Area(ULONG size, ULONG flags)
{
	VMMJmp(_Allocate_Device_CB_Area)
}
#pragma warning(default:4035)

extern int BSSstart, BSSend;

BOOL OnDeviceInit ()
{
	/* clear bss segment */
	memset (&BSSstart, 0, ((char *)&BSSend) - ((char *)&BSSstart));

	/* allocate VMM area for us to use to remember which memory
	   block we allocated for the client */
	cbmyVMarea = _Allocate_Device_CB_Area (sizeof (MEMORY_BLOCK), 0);
	if (!cbmyVMarea)
		return FALSE;

	return  TRUE;
}

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#pragma warning(disable:4035)
DWORD __declspec(naked) _Deallocate_Device_CB_Area(ULONG offset, ULONG flags)
{
	VMMJmp(_Deallocate_Device_CB_Area)
}
#pragma warning(default: 4035)

BOOL OnDeviceExit ()
{
	/* deallocate VMM area */
	if (cbmyVMarea)
		_Deallocate_Device_CB_Area (cbmyVMarea, 0);
	return TRUE;
}

BOOL EndPMApp2 (HVM hVM)
{
	/* A protected-mode application is ending.  Free its memory
	   block if it allocated one. */
	PMEMORY_BLOCK pmb = (PMEMORY_BLOCK) ((DWORD) hVM + cbmyVMarea);

	if (pmb->mb_addr)
		_PageFree((PVOID) pmb->mb_addr, 0);
	pmb->mb_addr = 0;
	pmb->mb_size = 0;
	return TRUE;
}


#pragma warning(disable: 4035)
DWORD __declspec(naked) _PageQuery(ULONG pbase,
	PMEMORY_BASIC_INFORMATION pmbi, 	ULONG cbmbi)
{
	VMMJmp(_PageQuery)
}

DWORD __declspec(naked) _PageReserve(ULONG page, ULONG npages, ULONG flags)
{
	VMMJmp(_PageReserve)
}

DWORD __declspec(naked) _PageCommit(ULONG page, ULONG npages, ULONG hpd, ULONG pagerdata, ULONG flags)
{
	VMMJmp(_PageCommit)
}

DWORD __declspec(naked) _PageDecommit(ULONG page, ULONG npages, ULONG flags)
{
	VMMJmp(_PageDecommit)
}

#if 0
DWORD __declspec(naked) _PageFree(ULONG page, ULONG flags)
{
	VMMJmp(_PageFree)
}
#endif

#define VTDI_DEVICE_ID			0x0488
Begin_Service_Table(VTDI, none)
#define VTDI_Service(X) Declare_Service(X, none)
VTDI_Service (VTDI_Get_Version)
VTDI_Service (VTDI_Start_Timer)
VTDI_Service (VTDI_Stop_Timer)
VTDI_Service (VTDI_Schedule_Event)
VTDI_Service (VTDI_Cancel_Event)
VTDI_Service (VTDI_Block)
VTDI_Service (VTDI_Signal)
VTDI_Service (VTDI_Register)
VTDI_Service (VTDI_Get_Info)
VTDI_Service (VTDI_Unload)
VTDI_Service (VTDI_Initialize)
VTDI_Service (VTDI_Register_UnloadProc)
VTDI_Service (VTDI_Register_LoadProc)
End_Service_Table(VTDI, none)

DWORD VTDI_Get_Version()
{
	VMMCall(VTDI_Get_Version)
	_asm jnc noerror
	_asm mov eax, 4321h
	_asm noerror:
}

DWORD __declspec(naked) VTDI_Get_Info(char *psz)
{
	VMMJmp(VTDI_Get_Info)
}


void PM_API (HVM, DWORD, PCRS);

void __declspec(naked) MMAP_PM_API ()
{
	_asm {
		push	ebp
		push	edi
		push	ebx
		call	PM_API
		pop		eax
		pop		eax
		pop		eax
		ret
	}
}

#pragma warning(default: 4035)


static char TCPname[] = "MSTCP";

MEMORY_BASIC_INFORMATION mbi;

void PM_API (HVM hVM, DWORD hThread, PCRS pRegs)
{
	DWORD res, clientbuf, len;
	PMEMORY_BLOCK pmb = (PMEMORY_BLOCK) ((DWORD) hVM + cbmyVMarea);

	switch (pRegs->Client_EAX) {
	case 0:				// GetVersion
		pRegs->Client_EAX = 0x0b00;
		pRegs->Client_EFlags &= ~CF_MASK;
		break;

	case 1:				// PageQuery
		res = _PageQuery (pRegs->Client_ESI, &mbi, sizeof(mbi));
		pRegs->Client_EBX = clientbuf = Client_Ptr_Flat(CLIENT_DS, CLIENT_DI);
		pRegs->Client_EDX = res;
	/*	if (res) { */
			len = (sizeof(mbi) > pRegs->Client_ECX) ? pRegs->Client_ECX : sizeof(mbi);
			memcpy ((PCHAR) clientbuf, &mbi, len);
			pRegs->Client_EAX = len;
			pRegs->Client_EDX = len;
			pRegs->Client_EFlags &= ~CF_MASK;
	/*	} else {
			pRegs->Client_EAX = 0;
			pRegs->Client_EFlags |= CF_MASK;
		}
	*/	break;
		
	case 2:				// GetCurrentContext
		pRegs->Client_EAX = _GetCurrentContext();
		pRegs->Client_EFlags &= ~CF_MASK;
		break;

	case 3:				// PageReserve
		if (pmb->mb_addr) {
			// already reserved an area
			pRegs->Client_EAX = 0xffffffff;
			pRegs->Client_EFlags |= CF_MASK;
			break;
		}
		pmb->mb_size = pRegs->Client_ECX << 12;
		pRegs->Client_EAX = _PageReserve(PR_SHARED, pRegs->Client_ECX, 0);
		if (pRegs->Client_EAX == 0xffffffff)
			pRegs->Client_EFlags |= CF_MASK;
		else {
			pmb->mb_addr = pRegs->Client_EAX;
			pRegs->Client_EFlags &= ~CF_MASK;
		}
		break;

	case 4:				// PageCommit
		if ((pRegs->Client_EDX > pmb->mb_size) ||
			(pRegs->Client_EDX + pRegs->Client_ECX > pmb->mb_size)) {
			pRegs->Client_EAX = 0x7777;
			pRegs->Client_EFlags |= CF_MASK;
			break;
		}
		pRegs->Client_EAX = _PageCommit(
			 (pmb->mb_addr + pRegs->Client_EDX) >> 12, 
			 pRegs->Client_ECX >> 12, 
			 PD_ZEROINIT, 0, PC_USER | PC_WRITEABLE);
		if (pRegs->Client_EAX == 0)
			pRegs->Client_EFlags |= CF_MASK;
		else
			pRegs->Client_EFlags &= ~CF_MASK;
		break;

	case 5:				// PageDecommit
		if ((pRegs->Client_EDX > pmb->mb_size) ||
			(pRegs->Client_EDX + pRegs->Client_ECX > pmb->mb_size)) {
			pRegs->Client_EAX = 0x7777;
			pRegs->Client_EFlags |= CF_MASK;
			break;
		}
		pRegs->Client_EAX = _PageDecommit(
			(pmb->mb_addr + pRegs->Client_EDX) >> 12,
			pRegs->Client_ECX >> 12, 0);
		if (pRegs->Client_EAX == 0)
			pRegs->Client_EFlags |= CF_MASK;
		else
			pRegs->Client_EFlags &= ~CF_MASK;
		break;

	case 6:				// PageFree
		if (pRegs->Client_EDX == 0)
			pRegs->Client_EDX = pmb->mb_addr;
		if (pRegs->Client_EDX != pmb->mb_addr) {
			pRegs->Client_EAX = 0x7777;
			pRegs->Client_EFlags |= CF_MASK;
			break;
		}
		pRegs->Client_EAX = _PageFree((PVOID) pRegs->Client_EDX, 0);
		if (pRegs->Client_EAX == 0)
			pRegs->Client_EFlags |= CF_MASK;
		else {
			pmb->mb_addr = 0;
			pmb->mb_size = 0;
			pRegs->Client_EFlags &= ~CF_MASK;
		}
		break;

	case 7:				// GetInfo -- get stored pmb info
		pRegs->Client_EDX = pmb->mb_addr;
		pRegs->Client_ECX = pmb->mb_size;
		pRegs->Client_EFlags &= ~CF_MASK;
		break;

	case 100:
		pRegs->Client_EAX = VTDI_Get_Version();
		break;

	case 101:
		pRegs->Client_EAX = VTDI_Get_Info(TCPname);
		break;

	default:
		pRegs->Client_EAX = 0x1234;
		pRegs->Client_EFlags |= CF_MASK;
		break;
	}
}
